/*
 * CocoWirrel 1.0.
 *
 * Un "resolvedor" del rompecabezas Cococrash "basado" en Wirrel, de
 * Thomer M. Gil. (http://www.thomer.com/wirrel/)
 *
 * Cambios hechos al cdigo fuente original (Wirrel 1.5):
 *
 * - Cambiado bool por _Bool
 * - Cambiado true por 1
 * - Cambiado false por 0
 * - Dentro de la funcin build he aadido "struct" al principio de la lnea
 *   que pone "piece_t *wd = &(pieces_list[j]);".
 *
 * Estos cambios son debidos a ligeras diferencias entre el compilador que
 * he usado (lcc-win32) y el usado por el autor (el de GNU, me parece).
 *
 * - Cambiada de lugar la llamada a la funcin init que aparece en la
 *   funcin main. Parece que era un error, ya que init machacaba valores
 *   establecidos al ejecutar read_pieces. En concreto el efecto notado era
 *   que no se tenan en cuenta las simetras de las piezas, produciendo
 *   muchas soluciones equivalentes.
 *
 * - La solucin muestra tambin la cadena de bits de cada pieza en su
 *   posicin inicial (no de acuerdo con la posicin en que se coloca).
 *
 * - Dentro de la funcin read_pieces, he cambiado el lmite inferior del
 *   bucle for principal de 0 a navailable [for (...i = 0...) cambiado por
 *   for (...i = navailable)]. Esto pemite usar varios ficheros de piezas.
 *
 * - He aadido dos arrays llamados piezas_que_ya_no_valen y
 *   caras_equivalentes_a_la_primera que se utilizan para evitar la
 *   produccin de algunas soluciones equivalentes. La idea es que en muchas
 *   figuras, muchas caras (llamo cara a cada una de las piezas que componen
 *   la figura, que en muchos casos pueden ser un trozo de cara en el
 *   sentido geomtrico) son "equivalentes", por simetra, a la primera. Es
 *   decir, que una vez probada una pieza en la primera cara, ya no tiene
 *   sentido probarla (como primera pieza) en otras equivalentes. Por
 *   ejemplo, tanto en el cubo 1x1x1 como en el 2x2x2, todas las caras son
 *   equivalentes en este sentido. Para este nuevo rasgo, en el fichero que
 *   define la figura (make file o recipe file, lo llama Gil) hay que poner
 *   como primera lnea una lista con los nmeros de las caras equivalentes
 *   a la primera, terminando con -1. Por ejemplo, en el caso del cubito
 *   1x1x1, la primera lnea debe ser 2 3 4 5 6 -1. Esto no est todava
 *   muy probado.
 *
 * - Cambiado el nombre a CocoWirrel y el nmero de versin a 1.0.
 *
 * - Traducidos los argumentos:
 *
 *   -h --help        ---> -? --ayuda (estos son en realidad ficticios)
 *   -d --description ---> -p --piezas
 *   -m --recipe_file ---> -f --figura
 *   -s --silent      ---> -s --silencioso
 *   -v --verbosity   ---> -l --locuaz
 *   -c --count       ---> -c --contar
 *
 * - Cambiado el argumento por omisin referido al fichero de figura. Era
 *   des/wir222.mak y ahora es figuras/cubo1x1x1.
 *
 * - Eliminado el argumento por omisin referido a fichero de piezas. Ahora
 *   es obligatorio especificar al menos uno.
 * 
 * - He suprimido la llamada a la funcin find_equals para que detecte
 *   soluciones con piezas iguales pero de distinto color.
 *
 * - Traducidos muchos mensajes.
 *
 * Rodolfo Valeiras Reina.
 *
 * -- RVR
 *
 */

/* vim: set wm=0 autowrite:
 *
 * wirrel.c -- Builds Happy Cubes (Wirrel Warrels)
 * Copyright (C) 1992-2002 Thomer M. Gil
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Thomer M. Gil (e-mail: [my first name]@[my last name].com)
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define VERSION                 "1.0" // cambiado, era 1.5 -- RVR

#define MAX_WIR_NAME_LEN       32	// max name of wirrel
#define NR_WIRS               180	// max # of wirrels
#define NBITS                  16	// # of bits on wirrel
#define BIT_SPECS               5	// how much info to ...Fit()
#define COUNT_STOP        1000000       // how often report?
#define NR_EQUALS              16	// how many equals in a set pieces?

//#define WIR_DES_FILE            "des/wirlist.des" suprimido, es obligatorio especificar un archivo de piezas
#define WIR_MAKE_FILE           "figuras/cubo1x1x1" //cambiado, era "mak/wir222.mak"
#define MAX_FILE_LEN            32	/* Max name of file */

#define BIT(n)  ((structure[FitBits[n].piece]->bit[structure[FitBits[n].piece]->pos] >> FitBits[n].bit) & 1)

/* Description of piece */
struct piece_t {
  char name[MAX_WIR_NAME_LEN];
  short bit[8];                 // bit pattern for all positions
  char symm[8];		        // symm[n] == m means pos n equal to pos m
  signed equals[NR_EQUALS];	// equals[] == m means this piece equals m
  _Bool busy;		        // is used in structure
  unsigned nequals;	        // How many equal pieces?
  char pos;		        // index into bit[]
};

/* This struct narrows down to one bit on one piece */
struct bit_t {
  short piece;                  /* What piece */
  char bit;                     /* What bit */
};

/* This struct contains a function pointer and an array with piece- and bit
 * numbers.
 */
struct recipe_t {
  _Bool valid;
  int (*Fits) (struct bit_t FitBits[BIT_SPECS],	/* Function to call */
	       struct piece_t * structure[NR_WIRS]);
  struct bit_t FitBits[BIT_SPECS];	/* Parameter to function */
};

unsigned nsolutions = 0;	// solutions found so far
unsigned ncomplete = 0;	        // number of pieces in complete structure
unsigned navailable = 0;	// number of available pieces
unsigned long int tries = 0;	// total combinations tried
unsigned counter = 1;	        //
int errno;
unsigned verbose = 0;           // verbosity
unsigned silence = 0;           // solution silence
char docount = 0;               // whether or not to count

struct piece_t pieces_list[NR_WIRS];
struct recipe_t recipe[NR_WIRS][NBITS];

int piezas_que_ya_no_valen[NR_WIRS]; // ndices de las piezas que ya no valen -- RVR
int caras_equivalentes_a_la_primera[NR_WIRS]; // ndices de las "caras" ... -- RVR


void
usage(char *argv[])
{
  printf("\
CocoWirrel %s\n\
Un resolvedor de Cococrash \"basado\" en Wirrel 1.5, de Thomer M. Gil\n\
\n\
Modo de uso: %s [opcin]...\n\
\n\
opciones:\n\
  -?, --ayuda                 muestra esta ayuda\n\
  -p, --piezas archivo        archivo contiene las descripciones de las piezas\n\
  -f, --figura archivo        archivo contiene la descripcin de la figura\n\
  -s, --silencioso            contina despus de dar una solucin\n\
  -c, --contar                da informacin sobre conteo\n\
  -l, --locuaz                aade locuacidad (puede usarse hasta tres veces)\n", VERSION, argv[0]);
}


void print_bits(short bits)
{
  for (unsigned i = 0; i < NBITS; i++)
    printf("%d", (bits & (1 << i)) != 0);
}


short
flip(short bits, unsigned position)
{
  if (verbose >= 3) {
    printf("Flipping from pos %d: ", position);
    print_bits(bits);
    printf("\n");
  }

  // turn right by shifting bits to left
  if (position != 3 && position != 7)
    return ((bits << 4)  & 0xfff0) | ((bits >> 12) & 0x000f);

  // flip piece
  return bits & (0x4000 | 0x40) | // bit 14 and 6 remain the same
    ((bits >> 2) & 0x2000)      | // bit 15 becomes bit 13
    ((bits << 2) & 0x8000)      | // bit 13 becomes bit 15
    ((bits >> 12) & 0x01)       | // bit 12 becomes bit 0
    ((bits >> 10) & 0x02)       | // bit 11 becomes bit 1
    ((bits >> 8) & 0x04)        | // bit 10 becomes bit 2
    ((bits >> 6) & 0x08)        | // bit 9 becomes bit 3
    ((bits >> 4) & 0x10)        | // bit 8 becomes bit 4
    ((bits >> 2) & 0x20)        | // bit 7 becomes bit 5
    ((bits << 2) & 0x80)        | // bit 5 becomes bit 7
    ((bits << 4) & 0x0100)      | // bit 4 becomes bit 8
    ((bits << 6) & 0x0200)      | // bit 3 becomes bit 9
    ((bits << 8) & 0x0400)      | // bit 2 becomes bit 10
    ((bits << 10) & 0x0800)     | // bit 1 becomes bit 11
    ((bits << 12) & 0x1000);      // bit 0 becomes bit 12
}



void
find_equals()
{
  for (unsigned i = 0; i < navailable; i++)
    for (unsigned j = i+1; j < navailable; j++)
      for (unsigned l = 0; l < 8; l++)
        if (pieces_list[i].bit[0] == pieces_list[j].bit[l]) {
	  pieces_list[i].equals[pieces_list[i].nequals++] = j;
	  pieces_list[j].equals[pieces_list[j].nequals++] = i;

	  /* Is NR_EQUALS big enough? */
	  if ((pieces_list[i].nequals == NR_EQUALS-1) ||
	      (pieces_list[j].nequals == NR_EQUALS-1)) {
	    fprintf(stderr, "Warning: increase NR_EQUALS!\n");
	  }

          if (verbose >= 2)
            printf("%s is equal to %s in position %d\n",
                  pieces_list[i].name, pieces_list[j].name, l);

	  // Not necessary to check it for other positions of j
	  break;
	}
}


void
init()
{
  int i, j;

  for (i = 0; i < NR_WIRS; i++) {
    pieces_list[i].busy = 0;//false;
    pieces_list[i].nequals = 0;

    piezas_que_ya_no_valen[i] = -1; // piezas numeradas desde el 0 -- RVR
    pieces_list[i].bit[0] = 0; // no parece hacer falta -- RVR

    for (j = 0; j < 8; j++)
      pieces_list[i].symm[j] = 0;
    for (j = 0; j < NR_EQUALS; j++)
      pieces_list[i].equals[j] = -1;
  }

  for (i = 0; i < NR_WIRS; i++)
    for (j = 0; j < NBITS; j++)
      recipe[i][j].valid = 0;//false;
}

void
build(struct piece_t *structure[NR_WIRS], unsigned i)
{
  unsigned nhist = 0;
  unsigned history[NR_WIRS];

  // Finished?
  if (i == ncomplete) {

    nsolutions++;

    if (silence >= 3)
      return;

    // Report a solution
    if (docount)
      printf("Solucion %d (encontrada en el intento %lu)\n", nsolutions, tries);
    else
      printf("Solucion %d\n", nsolutions);

    if (silence >= 2)
      return;

    for (unsigned j = 0; j < ncomplete; j++) {
      printf("%2d-%s: %d. ", j+1, structure[j]->name, structure[j]->pos+1);
      // -- RVR
      print_bits(structure[j]->bit[0]);//structure[j]->pos]);
      printf("\n");
      // -- RVR
    }

    if (silence < 1) {
      printf("Pulsa Intro para continuar\n");
      getchar();
    }

    printf("\n");
    return;
  }


  // go through all pieces and try to fulfill all requirements by calling
  // function pointers. all should return true.
  for (unsigned j = 0; j < navailable; j++) {
    // Only change first piece when there are more pieces available than
    // needed. Otherwise this is the point where we quit.
    if (!i && j && !(navailable > ncomplete))
      return;

    // -- RVR
    // Esto es para evitar soluciones equivalentes.
    // Ver nota de ms abajo.
    {
       int bb = 0;
       if (i) {
          for (int j = 0; j < ncomplete; j++) {
             if (caras_equivalentes_a_la_primera[j] == -1) break;
             if (caras_equivalentes_a_la_primera[j]-1 == i) {
                bb = 1;
                break;
             }
          }
       }
       int b = 0;
       for (int i = 0; i < navailable; i++) {
          if (piezas_que_ya_no_valen[i] == -1) break;
          if (piezas_que_ya_no_valen[i] == j) {
             b = 1;
             break;
          }
       }
       if (b && bb) continue;
    }
    // -- RVR

    struct piece_t *wd = &(pieces_list[j]);
    if (wd->busy)
      continue;

    // if an equally shaped piece has been tried already, skip this one
    if (wd->nequals) {
      unsigned useless = 0;
      for (unsigned j = 0; j < nhist; j++)
        for (unsigned k = 0; k < wd->nequals; k++)
          if (history[j] == (unsigned) wd->equals[k])
            useless = history[j];

      if (useless) {
        if (verbose >= 2)
          printf("No use to try %s in pos %d, because %s is the same!\n",
                wd->name, i, pieces_list[useless].name);
	continue;
      } else if (verbose >= 2)
	printf("Yes, equal to SOMETHING but first. Going ahead with %s in pos %d\n",
	   wd->name, i);
    }

    // remember that we tried piece j.
    history[nhist++] = j;

    // put piece in structure
    structure[i] = wd;

    // try piece in 8 different ways
    for (wd->pos = 0; wd->pos < 8; wd->pos++) {
      if (verbose >= 2)
        printf("Trying piece %s in pos %d, local pos %d\n", wd->name, i, wd->pos);

      if (wd->symm[wd->pos]) {
        if (verbose >= 2)
          printf("Skipping %s : %d\n", wd->name, wd->pos);
        continue;
      }

      if (docount) {
        tries++;
        if (counter == COUNT_STOP) {
          printf("Try %lu coming up...\n", tries);
          counter = 1;
        } else
          counter++;
      }

      // satisfy all set conditions
      _Bool fits = 1;//true;
      for (unsigned l = 0; l < NBITS; l++) {
        struct recipe_t *wl = &(recipe[i][l]);
        if (!wl->valid)
          break;

        if (!(*wl->Fits) (wl->FitBits, structure)) {
          fits = 0;//false;
          if (verbose >= 2)
            printf("Fault at %d\n", l);
          break;
        }
      }

      if (!fits)
        continue;

      // put piece in structure and descend
      if (verbose >= 2)
        printf("*** Fits: %s, %d, %d ***\n", wd->name, i, wd->pos);
      wd->busy = 1;//true;
      build(structure, i+1);

      // don't try first piece in other positions
      if (!i)
        break;
    }

    // Take piece out of structure
    wd->pos = 0;
    wd->busy = 0;//false;

    // -- RVR
    // Esto es para evitar soluciones equivalentes.
    // Cuando una pieza (j) se ha probado en la primera cara (i == 0) con todas las dems
    // piezas, ya no ser necesario probarla en otra cara equivalente.
    if (i == 0) {
       int iepqynv;
       for (int i = 0; i < navailable; i++) {
          if (piezas_que_ya_no_valen[i] == -1) {
             iepqynv = i;
             break;
          }
       }
       piezas_que_ya_no_valen[iepqynv] = j;
    }
    // -- RVR

    if (verbose >= 2)
      printf("Took out piece %s from position %d.\n", wd->name, i);
  }
}



int
SimpleFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  return BIT(0)+BIT(1) == 1;
}


int
OpenFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  /* Return true if either * sum in corner is 1 * sum in corner is 0 AND (one
     of the bits next to corner is open) */
  register int sum;

  sum = BIT(0)+BIT(1);
  return (sum == 1 || ((!sum) && (!BIT(2) || !BIT(3))));
}

int
CornerFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  return (BIT(0)+BIT(1)+BIT(2) == 1);
}


int
QuadFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  return (BIT(0)+BIT(1)+BIT(2)+BIT(3) == 1);
}


int
DoubleFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  return (BIT(0)+BIT(1) <= 1);
}


int
TripleFit(struct bit_t FitBits[BIT_SPECS], struct piece_t *structure[NR_WIRS])
{
  /* Return true if three bits are together 1 OR * three bits are together 0
     and empty bit is not closed in. */
  register int sum;

  sum = BIT(0)+BIT(1)+BIT(2);
  return (sum == 1 || ((!sum) && (!BIT(3) || !BIT(4))));
}



int
read_pieces(char File[MAX_FILE_LEN])
{
  FILE *fp;

  /* Open file */
  if ((fp = fopen(File, "r")) == NULL) {
    fprintf(stderr, "No puedo abrir el archivo %s. Errno = %d.\n", File, errno);
    return 0;
  }

  // Read one line from file
  // -- RVR
  // Pequeo cambio para que se puedan usar varios ficheros de piezas
  for (unsigned i = navailable; i < NR_WIRS && !feof(fp); i++) {
  // --RVR
//  for (unsigned i = 0; i < NR_WIRS && !feof(fp); i++) { // <-- estaba as
    struct piece_t *wd = &(pieces_list[i]);

    // Read name
    int r = fscanf(fp, "%s", wd->name);
    if (feof(fp))
      return i;
    if (!r) {
      fprintf(stderr, "No puedo leer nombre en el archivo de piezas\n");
      return 0;
    }

    // Read bits
    unsigned bit;
    for (unsigned j = 0; j < NBITS; j++) {
      if (!fscanf(fp, "%1u", &bit)) {
	fprintf(stderr, "No puedo leer bits en el archivo de piezas\n");
	return 0;
      }
      wd->bit[0] |= (bit << j);
    }

    if (verbose >= 2) {
      printf("%s: ", wd->name);
      print_bits(wd->bit[0]);
      printf("\n");
    }

    // Generate all possible positions for this piece
    for (unsigned j = 1; j < 8; j++)
      wd->bit[j] = flip(wd->bit[j-1], j-1);

    // Start position
    wd->pos = 0;

    // Do a symmetry check
    for (unsigned j = 1; j < 8; j++) {
      for (signed k = j-1; k >= 0; k--) {
        if (wd->bit[j] == wd->bit[k]) {
          wd->symm[j] = 1;
          break;
        }
      }
    }

    navailable++;
  }

// aqu nunca llega

  return navailable;
}


int
read_recipe(char File[MAX_FILE_LEN])
{
  FILE *fp;

  /* Open file */
  if ((fp = fopen(File, "r")) == NULL) {
    fprintf(stderr, "No puedo abrir el archivo %s. Errno = %d.\n", File, errno);
    return 0;
  }
  
  // -- RVR  Para evitar soluciones equivalentes. Al principio del fichero que describe
  // la figura debe haber una lista de los nmeros de las caras (piezas en la figura)
  // equivalentes a la primera. La lista debe terminar con -1.
  int i = 0;
  while (1)
     if (
           !fscanf(fp, "%d", &caras_equivalentes_a_la_primera[i]) ||
           (caras_equivalentes_a_la_primera[i++] == -1)
     ) break;
  // -- RVR

  /* Read first piece number */
  unsigned piece_no;
  if (!fscanf(fp, "%u:", &piece_no)) {
    fprintf(stderr, "No puedo leer el numero de cara en el archivo de figura\n");
    return 0;
  }
  ncomplete++;

  /* FOR every piece, read the description how to fit it in cube */
  for (unsigned i = 0; i < navailable && !feof(fp); i++) {
    if (verbose >= 2)
      printf("Reading piece %d\n", ncomplete);

    for (unsigned j = 0; j < NBITS; j++) {
      int r = fscanf(fp, " %u:", &piece_no);
      if (feof(fp))
	break;

      /* Isn't this start of new piece? Did the %d work? */
      if (r) {
	ncomplete++;
	break;
      }

      /* %d didn't work. It is a letter. Read it */
      char func_type;
      r = fscanf(fp, " %c:", &func_type);

      if (!r) {
	fprintf(stderr, "No puedo leer tipo de funcion en el archivo de figura: %d\n", r);
	return 0;
      }

      unsigned specs;
      switch (func_type) {
      case ('S'):
	recipe[i][j].Fits = SimpleFit;
	specs = 2;
	break;
      case ('O'):
	recipe[i][j].Fits = OpenFit;
	specs = 4;
	break;
      case ('C'):
	recipe[i][j].Fits = CornerFit;
	specs = 3;
	break;
      case ('Q'):
	recipe[i][j].Fits = QuadFit;
	specs = 4;
	break;
      case ('D'):
	recipe[i][j].Fits = DoubleFit;
	specs = 2;
	break;
      case ('T'):
	recipe[i][j].Fits = TripleFit;
	specs = 5;
	break;
      default:
        fprintf(stderr, "Tipo desconocido %c\n", func_type);
        exit(-1);
      }

      /* specs says how many n.n we expect */
      for (unsigned k = 0; k < specs; k++) {
        unsigned p, b;
	if (!fscanf(fp, " %u.%u", &p, &b)) {
	  fprintf(stderr, "No puedo leer n.n del archivo de figura en la cara %d\n",
		  ncomplete+1);
	  return 0;
	}

        if (verbose >= 2)
          printf("p = %d, b = %d\n", p, b);

	recipe[i][j].FitBits[k].piece = p-1;
	recipe[i][j].FitBits[k].bit = b-1;

	// read a "-" when necessary
	if (k < specs-1) {
          char dummy;
	  if (!fscanf(fp, " %c", &dummy)) {
	    fprintf(stderr, "No puedo leer del archivo de figura\n");
	    return 0;
	  }
	}
      }
      recipe[i][j].valid = 1;//true;
    }
  }

  return ncomplete;
}



int
main(int argc, char *argv[])
{
  struct piece_t *structure[NR_WIRS];
  _Bool readpieces = 0;//false;
  char recipe_file[MAX_FILE_LEN] = WIR_MAKE_FILE;
  int i;

  init(); // estaba abajo  -- RVR
    
  /* Parse arguments */

  for (i = 1; argv[i]; i++)
    if (!strcmp("-p", argv[i]) || !strcmp("--piezas", argv[i])) {
      if (!read_pieces(argv[++i])) {
        fprintf(stderr, "No encuentro el archivo de piezas.\n");
        exit(-1);
      }
      readpieces = 1;//true;

    } else if (!strcmp("-f", argv[i]) || !strcmp("--figura", argv[i]))
      strcpy(recipe_file, argv[++i]);
    else if (!strcmp("-s", argv[i]) || !strcmp("--silencioso", argv[i]))
      silence++;
    else if (!strcmp("-c", argv[i]) || !strcmp("--contar", argv[i]))
      docount++;
    else if (!strcmp("-l", argv[i]) || !strcmp("--locuaz", argv[i]))
      verbose++;
    else {
      usage(argv);
      exit(-1);
    }

//  init();

  if(!readpieces) {
      usage(argv);
      exit(-1);
  }

  // read recipe
  if (!read_recipe(recipe_file)) {
    fprintf(stderr, "No encuentro el archivo de figura.\n");
    exit(-1);
  }

//  find_equals(); -- RVR

  if (verbose >= 1)
    printf("ncomplete = %d\n", ncomplete);

  // Are there enough pieces?
  if (ncomplete > navailable)
    fprintf(stderr, "No hay suficientes piezas para construir la figura.\n");

  // Get the ball rolling
  build(structure, 0);

  return 0;
}
