{****************************************************************************

 Programa complementario de CocoWirrel.

 Usos:

    Ver directamente una pieza
       coco 0010001000100101
    Ver todas las piezas de un fichero de piezas
       coco -p piezas\azul
    Ver una solucin pieza a pieza (modo desarrollo)
       coco -d sol\azul.sol 1
    Ver una solucin en modo figura
       coco -s sol\azul.sol 1 cubo1x1x1.map
       coco -s sol\azul.sol 1
    Filtrar un fichero de soluciones
       coco -f sol\azul.sol cubo1x1x1.map
       coco -f sol\azul.sol
    Extraer soluciones de un fichero de soluciones
       coco -e sol\ZVANRM.sol azul 3 verde 3
       coco -es sol\ZVANRM.sol azul +3

 Se ha compilado en Windows XP usando Free Pascal 1.0.6.

 Creado el lunes, 24 de marzo de 2003
 ltima modificacin: sbado, lunes, 12 de mayo de 2003


 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 ****************************************************************************}


uses Crt;

const
   MaxNPiezas = 36;
   NPos = 8;
   NDirec = 5;

   ARRIBA = 1;
   ABAJO = 2;
   IZQUIERDA = 3;
   DERECHA = 4;
   FRENTE = 5;

type
   TPieza = 0..MaxNPiezas;
   TPos = 1..NPos;
   TDirec = 1..NDirec;
   TNodo = record
      Pieza : TPieza;
      Pos : TPos;
   end;
   TCocoSim = Word;
   TCoco = record
      Nombre : String;
      Posi : TPos;
      Desc : array[TPos] of TCocoSim;
   end;
   TMapa = array[TPieza, TPos, TDirec] of TNodo;
   TSolucion = array[TPieza] of TCoco;
   TStBit = '0'..'1';
   TStWord = array[1..16] of TStBit;

var
   Mapa : TMapa;

const
   NPiezas : Byte = 6;


{ -------------------------------------------------------------------------}


var
   Linea : String;


{
   Lee un nmero de la cadena Linea. Es auxiliar de LeeMapa.
   En caso de error devuelve -1.
}
function LeeNumero : Integer;
const
   i : Word = 1;
   j : Word = 1;
var
   v, CodErr : Word;
label
   FinDeLinea;
begin
   if i > Length(Linea) then goto FinDeLinea;
   while (Linea[i] < '0') or (Linea[i] > '9') do begin
      if i = Length(Linea) then goto FinDeLinea;
      Inc(i);
   end;
   j := i;
   while (Linea[j] >= '0') and (Linea[j] <= '9') do begin
      Inc(j);
      if j > Length(Linea) then break;
   end;
   Val(Copy(Linea, i, j-i), v, CodErr);
   i := j;
   Exit(v);
   FinDeLinea:
   i := 1;
   j := 1;
   Exit(-1);
end;


{
   Lee un fichero de mapa.
   Valor devuelto :
      0 : 0 sin error
      1 : No se encuentra el fichero
      2 : Hay errores en el fichero
}
function LeeMapa(var M : TMapa; NomFich : String) : Byte;
var
   FMapa : Text;
   i, Max : Byte;
   v1, v2 : Integer;
begin
   Assign(FMapa, NomFich);
   {$I-}
   Reset(FMapa);
   {$I+}
   if IOResult <> 0 then Exit(1);
   Max := 0;
   while not Eof(FMapa) do begin
      ReadLn(FMapa, Linea);
      if (Linea = '') or (Linea[1] = ';') then Continue;
      v1 := LeeNumero; if v1 = -1 then Exit(2); if v1 > Max then Max := v1;
      v2 := LeeNumero; if v1 = -1 then Exit(2);
      for i := 1 to NDirec do begin
         M[v1, v2, i].Pieza := LeeNumero;
         M[v1, v2, i].Pos := LeeNumero;
      end;
      v1 := LeeNumero; // para que LeeNumero detecte el fin de lnea
   end;
   Close(FMapa);
   NPiezas := Max;
   LeeMapa := 0;
end;


{
   Devuelve el nodo que resulta de ir desde A en la direccin D.
}
function Mueve(A : TNodo; D : TDirec) : TNodo;
begin
   Mueve := Mapa[A.Pieza, A.Pos, D];
end;


const
   PotenciasDeDos : array[0..15] of Word = (
        1,   2,    4,    8,   16,   32,    64,   128,
      256, 512, 1024, 2048, 4096, 8192, 16384, 32768
   );

{
   Pone a 1 o a 0 el bit nmero Pos de W.
}
procedure PoneBit(var W: TCocoSim; Bit: Boolean {True = 1, False = 0}; Pos: Byte {0..15});
begin
   if Bit
      then W := W or PotenciasDeDos[Pos]
      else W := W and not PotenciasDeDos[Pos];
end;


{
   Devuelve el valor normal (un Word) que corresponde a la cadena binaria Cad.
}
function LeeCoco(Cad : TStWord) : TCocoSim;
var
   i : Byte;
   C : TCocoSim;
begin
   for i := 1 to 16 do PoneBit(C, Cad[i] = '1', 16-i);
   LeeCoco := C;
end;


{
   Devuelve True si el bit nmero i de C es 1 y False el caso contrario.
}
function Bit(C : TCocoSim; i : Byte) : Boolean;
begin
   Bit := C and PotenciasDeDos[i] > 0;
end;


{
   Devuelve el string binario que corresponde a C.
}
function EscribeCoco(C : TCocoSim) : TStWord;
var
   i : Byte;
   S : TStWord;
   B : TStBit;
begin
   for i := 1 to 16 do begin
      if Bit(C, i-1) then B := '1' else B := '0';
      S[17-i] := B;
   end;
   EscribeCoco := S;
end;


{
   Devuelve la pieza C en la posicin P. C est en la posicin 1.
}
function PonPosCoco(C : TCocoSim; P : TPos) : TCocoSim;
var
   i, d : Byte;
   CPos : TCocoSim;
begin
   if P = 1 then Exit(C);
   if P > 4 then begin
      for i := 0 to 14 do PoneBit(CPos, Bit(C, 14-i), i);
      PoneBit(CPos, Bit(C, 15), 15);
   end;
   if P = 5 then Exit(CPos);
   if P < 5 then
      d := (P-1) * 4
   else begin
      C := CPos;
      d := (P-5) * 4;
   end;
   for i := 0 to 15 do PoneBit(CPos, Bit(C, (i+d) mod 16), i);
   PonPosCoco := CPos;
end;


{
   Rellena lo que falta de la variable Sol, a partir de la primera posicin
   de cada pieza.
}
procedure ExtiendeSol(var Sol : TSolucion);
var
   i : TPieza;
   j : TPos;
begin
   for i := 1 to NPiezas do
      with Sol[i] do
         for j := 2 to NPos do
            Desc[j] := PonPosCoco(Desc[1], j);
end;


{
   Lee una solucin (la que toca) de un fichero de soluciones abierto.
   Valor devuelto :
      0 : 0 sin error
      2 : No hay ms soluciones en el fichero
      3 : Superado el nmero mximo de piezas
}
function LeeSol(var Sol : TSolucion; var FichSol : Text) : Byte;
var
   L : String;
   Cod : Word;
   i, j, k : Byte;
begin
   if Eof(FichSol) then Exit(2);
   repeat
      ReadLn(FichSol, L);
   until (L <> '') and(Pos('Solucion', L) = 0);
   i := 1;
   repeat
      j := Pos('-', L)+1;
      k := Pos(':', L);
      Sol[i].Nombre := Copy(L, j, k-j);
      j := k+2;
      k := Pos('. ', L);
      Val(Copy(L, j, k-j), Sol[i].Posi, Cod);
      j := k+2;
      k := Length(L)+1;
      Sol[i].Desc[1] := LeeCoco(Copy(L, j, k-j));
      ReadLn(FichSol, L);
      if L = '' then Break;
      if i = MaxNPiezas then Exit(3);
      Inc(i);
   until False;
   LeeSol := 0;
end;


{
   Lee la solucin nmero n del fichero de soluciones llamado NomFich.
   Valor devuelto :
      0 : 0 sin error
      1 : No se encuentra el fichero
      2 : No hay tantas soluciones en el fichero
      3 : Superado el nmero mximo de piezas
}
function LeeSol(var Sol : TSolucion; NomFich : String; n : Word) : Byte;
var
   FichSol : Text;
   L : String;
   i : Word;
   vd : Byte;
begin
   Assign(FichSol, NomFich);
   {$I-}
   Reset(FichSol);
   {$I+}
   if IOResult <> 0 then Exit(1);
   i := 0;
   repeat
      if Eof(FichSol) then begin
         Close(FichSol);
         Exit(2);
      end;
      ReadLn(FichSol, L);
      if Pos('Solucion', L) > 0 then Inc(i);
      if i = n then vd := LeeSol(Sol, FichSol);
   until i = n;
   Close(FichSol);
   LeeSol := vd;
end;


{
   Muestra una pieza de forma grfica, pero en modo de texto, usando lneas
   y cruces.
}
procedure VerCoco(Cadena : TStWord);
const
   X0 = 30;
   Y0 = 8;
   Lado = 5;
var
   i, x, y : Byte;
   M : Array[0..Lado+1, 0..Lado+1] of Boolean;
begin
   // Llena la matriz M
   for y := 0 to Lado+1 do begin
      for x := 0 to Lado+1 do begin
         M[x, y] := (x > 1) and (x < Lado) and (y > 1) and (y < Lado);
      end;
   end;
   for i := 1 to Lado-1 do begin
      M[i, 1] := Cadena[i] = '1';
   end;
   for i := 1 to Lado-1 do begin
      M[Lado, i] := Cadena[i+Lado-1] = '1';
   end;
   for i := 1 to Lado-1 do begin
      M[Lado+1-i, Lado] := Cadena[i+2*(Lado-1)] = '1';
   end;
   for i := 1 to Lado-1 do begin
      M[1, Lado+1-i] := Cadena[i+3*(Lado-1)] = '1';
   end;
   // Borra el espacio donde va a dibujar
   for i := 1 to 2*Lado+2 do begin
      GotoXY(X0, Y0+i-1);
      Write(StringOfChar(' ', 4*Lado+1));
   end;
   // Dibuja las lneas verticales
   for y := 1 to Lado do begin
      for x := 0 to Lado do begin
         if M[x, y] xor M[x+1, y] then begin
            GotoXY(X0+4*x, Y0+2*(y-1)+1);
            Write('|');
         end;
      end;
   end;
   // Dibuja las lneas horizontales
   for x := 1 to Lado do begin
      for y := 0 to Lado do begin
         if M[x, y] xor M[x, y+1] then begin
            GotoXY(X0+4*(x-1)+1, Y0+2*y);
            Write('---');
         end;
      end;
   end;
   // Dibuja las cruces
   for x := 0 to Lado do begin
      for y := 0 to Lado do begin
         if
            (M[x, y] or M[x+1, y] or M[x+1, y+1] or M[x, y+1]) and
            not (M[x, y] and M[x+1, y] and M[x+1, y+1] and M[x, y+1])
         then begin
            GotoXY(X0+4*x, Y0+2*y);
            Write('+');
         end;
      end;
   end;
end;


{
   Devuelve la posicin en que se transforma Pos despus de aplicarle el
   movimiento que corresponde a t ('d', 'i', 'h' o 'v').
}
function CambiaPos(Pos : TPos; t : Char) : TPos;
begin
   if (t = 'd') or (t = 'D') then
      if Pos < 5 then CambiaPos := (Pos mod 4)+1 else CambiaPos := ((Pos-4) mod 4)+5 else
   if (t = 'i') or (t = 'I') then
      if Pos < 5 then CambiaPos := ((Pos+2) mod 4)+1 else CambiaPos := ((Pos-2) mod 4)+5 else
   if (t = 'h') or (t = 'H') then
      if (Pos = 1) or (Pos = 3) then CambiaPos := Pos + 5 else
      if (Pos = 6) or (Pos = 8) then CambiaPos := Pos - 5 else
      if (Pos = 2) or (Pos = 4) then CambiaPos := Pos + 3 else
      if (Pos = 5) or (Pos = 7) then CambiaPos := Pos - 3 else else
   if (t = 'v') or (t = 'V') then
      if (Pos = 1) then CambiaPos := 8 else
      if (Pos = 2) then CambiaPos := 7 else
      if (Pos = 3) then CambiaPos := 6 else
      if (Pos = 4) then CambiaPos := 5 else
      if (Pos = 5) then CambiaPos := 4 else
      if (Pos = 6) then CambiaPos := 3 else
      if (Pos = 7) then CambiaPos := 2 else
      if (Pos = 8) then CambiaPos := 1;
end;


{
   Muestra una pieza en cualquier posicin.
   Devuelve la tecla pulsada para salir (s, S, q o Q).
}
function MuestraCoco(Cadena : TStWord; Nombre : String; Pos : TPos; N : TPieza) : Char;
var
   t : Char;
   W0, W : TCocoSim;
begin
   ClrScr;
   CursorOff;
   Write('Pulsa q para terminar');
   W0 := LeeCoco(Cadena);
   Cadena := EscribeCoco(PonPosCoco(W0, Pos));
   repeat
      GotoXY(1, 2);
      if N > 0 then Write(N, '-');
      if Length(Nombre) > 0 then Write(Nombre, ' ');
      Write(Pos, '. ', Cadena);
      VerCoco(Cadena);
      repeat
         t := ReadKey;
         if t = #0 then begin
            t := ReadKey;
            if t = 'K' then t := 'i' else
            if t = 'M' then t := 'd' else
            if (t = 'H') or (t = 'P') then t := 'v';
         end;
         if (t>'0') and (t<'9') or
            (t='q') or (t='Q') or
            (t='s') or (t='S') or
            (t='i') or (t='I') or
            (t='d') or (t='D') or
            (t='h') or (t='H') or
            (t='v') or (t='V') then Break;
      until False;
      if (t > '0') and (t < '9') then Pos := Ord(t)-Ord('0') else
      Pos := CambiaPos(Pos, t);
      W := PonPosCoco(W0, Pos);
      Cadena := EscribeCoco(W);
   until (t = 'q') or (t = 'Q') or (t = 's') or (t = 'S');
   CursorOn;
   GotoXY(1, 24);
   MuestraCoco := t;
end;


{
   Un coco est colocado como determinada pieza (N) de una figura en la posicin Pos1
   (posicin en N-1). Devuelve la posicin de ese coco que le corresponde al nodo
   N-Pos.
}
function Posicion(Pos, Pos1 : TPos) : TPos;
var
   p : TPos;
begin
   p := Pos1;
   if Pos > 1 then p := CambiaPos(p, 'd');
   if Pos > 2 then p := CambiaPos(p, 'd');
   if Pos > 3 then p := CambiaPos(p, 'd');
   if Pos > 4 then p := CambiaPos(p, 'v');
   if Pos > 5 then p := CambiaPos(p, 'd');
   if Pos > 6 then p := CambiaPos(p, 'd');
   if Pos > 7 then p := CambiaPos(p, 'd');
   Posicion := p;
end;


{
   Muestra una solucin. Muestra una pieza cada vez, permitiendo "navegar" por
   la figura usando las teclas de direccin. Las instrucciones para navegar dependen
   de la variable Mapa.
}
procedure VerSol(Sol : TSolucion);
var
   Actual : TNodo;
   t : Char;
   d : TDirec;
   p : TPos;
begin
   Actual.Pieza := 1;
   Actual.Pos := 1;
   ClrScr;
   CursorOff;
   Write('Pulsa q para terminar');
   repeat
      with Actual do begin
         p := Posicion(Pos, Sol[Pieza].Posi);
         VerCoco(EscribeCoco(Sol[Pieza].Desc[p]));
         GotoXY(1, 2);
         WriteLn('                                                            ');
         GotoXY(1, 2);
         with Sol[Pieza] do
            WriteLn(Pieza, '-', Nombre, ' ', Pos, '. ', EscribeCoco(Desc[Pos]));
      end;
      t := ReadKey;
      if (t = 'q') or (t = 'Q') then Break;
      if t = 'A' then t := 'a' else
      if t = 'B' then t := 'b' else
      if t = 'I' then t := 'i' else
      if t = 'D' then t := 'd';
      if t = #0 then begin
         t := ReadKey;
         if t = 'H' then t := 'a' else
         if t = 'P' then t := 'b' else
         if t = 'K' then t := 'i' else
         if t = 'M' then t := 'd';
      end;
      if not (t in ['a', 'b', 'i', 'd', 'f']) then Continue;
      if t = 'a' then d := ARRIBA    else
      if t = 'b' then d := ABAJO     else
      if t = 'i' then d := IZQUIERDA else
      if t = 'd' then d := DERECHA   else
      if t = 'f' then d := FRENTE;
      Actual := Mueve(Actual, d);
   until t = 'q';
   CursorOn;
   GotoXY(1, 24);
end;

{
   Compara dos soluciones para ver si son equivalentes.
   Valor devuelto :
      0 : Soluciones equivalentes
      1 : No usan las mismas piezas
      2 : Usan las mismas piezas pero no son equivalentes
   *** De momento slo es vlida para cubos 1x1x1 ***
}
function SolEquiv(S1, S2 : TSolucion) : Byte;
var
   i, j : Byte;
   PosEq : array[0..NPos] of Byte;
   b : Boolean;
   N1, N2, N1Bis, N2Bis : TNodo;
begin
   for i := 1 to NPiezas do begin
      for j := 1 to NPiezas do begin
         b := (S1[i].Nombre = S2[j].Nombre);
         if (i = 1) and b then N2.Pieza := j; // localiza en S2 la primera pieza de S1
         if b then Break;
      end;
      if not b then Exit(1);
   end;
   for i := 1 to NPiezas do begin
      for j := 1 to NPiezas do begin
         b := (S2[i].Nombre = S1[j].Nombre);
         if b then Break;
      end;
      if not b then Exit(1);
   end;
   N1.Pieza := 1;
   N1.Pos := 1;
   FillByte(PosEq, NPos+1, 0);
   for i := 1 to NPos do begin
      if S2[N2.Pieza].Desc[Posicion(i, S2[N2.Pieza].Posi)] = S1[1].Desc[S1[1].Posi] then begin
         Inc(PosEq[0]);
         PosEq[PosEq[0]] := i;
      end;
   end;
   for i := 1 to PosEq[0] do begin
      N2.Pos := PosEq[i];
      for j := 1 to NDirec do begin
         N1Bis := Mueve(N1, j);
         N2Bis := Mueve(N2, j);
         b :=
            (S1[N1Bis.Pieza].Nombre <> S2[N2Bis.Pieza].Nombre) or
            (
               S1[N1Bis.Pieza].Desc[Posicion(N1Bis.Pos, S1[N1Bis.Pieza].Posi)] <>
               S2[N2Bis.Pieza].Desc[Posicion(N2Bis.Pos, S2[N2Bis.Pieza].Posi)]
            );
         if b then Break;
      end;
      if not b then Exit(0);
   end;
   SolEquiv := 2;
end;


{
   Busca una solucin en un fichero de soluciones.
   Devuelve el nmero de la solucin o:
      0 : No se encuentra la solucin
      -1 : Fichero no encontrado
      -3 : Rebasado el nmero mximo de piezas
}

function BuscaSol(Sol : TSolucion; NomFich : String) : Integer;
var
   i : Word;
   S : TSolucion;
   FichSol : Text;
begin
   Assign(FichSol, NomFich);
   {$I-}
   Reset(FichSol);
   {$I+}
   if IOResult <> 0 then Exit(-1);
   i := 0;
   repeat
      Inc(i);
      case LeeSol(S, FichSol) of
         1 : Exit(-1); // No se encuentra el fichero
         3 : Exit(-3); // Rebasado el nmero mximo de piezas
         2 : Exit(0);  // No est la solucin
      end;
      ExtiendeSol(S);
      if SolEquiv(S, Sol) = 0 then Exit(i);
   until False;
   Close(FichSol);
   BuscaSol := 0;
end;


{
   Filtra un fichero de soluciones buscando soluciones equivalentes.
   Crea un fichero con el mismo nombre y la extensin .filtrado que
   contiene slo soluciones no equivalentes.
   Devuelve:
      0 : Sin problemas
      1 : No encuentra el fichero
      2 : Se rebasa el lmite de soluciones (MaxSolUni)
      3 : Superado el nmero mximo de piezas
      4 : El fichero no tiene ninguna solucin
}
function FiltraSol(NomFich : String) : Byte;
const
   MaxSolUni = 7000; // Nmero mximo de soluciones nicas; es un valor arbitrario
type
   Bicho = Array[1..MaxSolUni] of TSolucion;
var
   F, G, FichSol : Text;
   NumSol, i, j : Word;
   ArrSol : ^Bicho;
   Indices : Array[1..MaxSolUni] of Word;
   Sol : TSolucion;
   b : Boolean;
begin
   Assign(F, ''); // output estndar, para poder redirigir
   Rewrite(F);
   Assign(G, NomFich + '.filtrado');
   Rewrite(G);
   Assign(FichSol, NomFich);
   {$I-}
   Reset(FichSol);
   {$I+}
   if IOResult <> 0 then Exit(1);
   New(ArrSol);
   case LeeSol(ArrSol^[1], FichSol) of
      3 : begin Dispose(ArrSol); Close(FichSol); Exit(3); end;
      2 : begin Dispose(ArrSol); Close(FichSol); Exit(4); end;
   end;
   ExtiendeSol(ArrSol^[1]);
   NumSol := 1;
   Indices[1] := 1;
   WriteLn(G, 'Solucion 1');
   for j := 1 to NPiezas do
      WriteLn(G, j:2, '-', ArrSol^[1, j].Nombre, ': ', ArrSol^[1, j].Posi, '. ',
         EscribeCoco(ArrSol^[1, j].Desc[1]));
   WriteLn(G);
   i := 1;
   repeat
      Inc(i);
      if LeeSol(Sol, FichSol) > 0 then Break; // fin del fichero
      ExtiendeSol(Sol);
      for j := 1 to NumSol do begin
         b := (SolEquiv(Sol, ArrSol^[j]) = 0);
         if b then Break;
      end;
      if b then begin
         WriteLn(F, 'La solucion ', i, ' es equivalente a la ', Indices[j]);
         Continue;
      end;
      if NumSol = MaxSolUni then begin
         Dispose(ArrSol);
         Close(F);
         Close(G);
         Close(FichSol);
         Exit(2);
      end;
      Inc(NumSol);
      ArrSol^[NumSol] := Sol;
      Indices[NumSol] := i;
      if i <> NumSol then WriteLn(F, 'La solucion ', i, ' renumerada como ', NumSol);
      WriteLn(G, 'Solucion ', NumSol);
      for j := 1 to NPiezas do
         WriteLn(G, j:2, '-', Sol[j].Nombre, ': ', Sol[j].Posi, '. ', EscribeCoco(Sol[j].Desc[1]));
      WriteLn(G);
   until False;
   Close(G);
   if i = (NumSol + 1) then begin
      Erase(G);
      WriteLn(F, 'No se ha eliminado ninguna solucion');
   end else
      WriteLn(F, 'Se ha creado el fichero ', NomFich + '.filtrado con ', NumSol, ' soluciones');
   Close(F);
   Close(FichSol);
   Dispose(ArrSol);
   FiltraSol := 0;
end;


function ByteNum(var CualquierCosa; Num : Longint) : Byte;
var
   Tabla : Array[1..1] of Byte absolute CualquierCosa;
begin
   ByteNum := Tabla[Num];
end;


type
   TPatron = Array[1..MaxNPiezas] of Record
      Color : String;
      Cantidad : Byte;
      Op : Char; // +, - o =
   end;

{
   Extrae de un fichero de soluciones aquellas que cumplen con un determinado
   patrn de colores, sin importar el orden. Por ejemplo, 3 piezas azules y
   3 amarillas. Si Silencio es True no se muestran las soluciones.
   Valor devuelto:
      >= 0 : Nmero de soluciones que extrae
      -1 : No se encuentra el fichero
      -3 : Superado el nmero mximo de piezas
}
function ExtraeSol(NomFich : String; Patron : TPatron; Silencio : Boolean) : Integer;
var
   NumSol, quesol : Word;
   Sol : TSolucion;
   vd, i, j, LonPatron : Byte;
   b : Boolean;
   Tabla : Array[TPieza] of Byte;
   F, FichSol : Text;
begin
   Assign(F, '');
   Rewrite(F);
   Assign(FichSol, NomFich);
   {$I-}
   Reset(FichSol);
   {$I+}
   if IOResult <> 0 then Exit(-1);
   for i := 1 to NPiezas do begin
      if i = NPiezas then LonPatron := i else
      if ByteNum(Patron[i], 1) = 0 then LonPatron := i-1
      else Continue;
      Break;
   end;
   NumSol := 0;
   quesol := 0;
   repeat
      Inc(quesol);
      vd := LeeSol(Sol, FichSol);
      case vd of
         3 : Exit(-3);
         2 : Break;
      end;
      FillByte(Tabla, MaxNPiezas+1, 0);
      for i := 1 to NPiezas do
         for j := 1 to LonPatron do
            if Pos(Patron[j].Color, Sol[i].Nombre) > 0 then Inc(Tabla[j]);
      b := True;
      for j := 1 to LonPatron do
         case Patron[j].Op of
            '=' : b := b and (Tabla[j] = Patron[j].Cantidad);
            '+' : b := b and (Tabla[j] >= Patron[j].Cantidad);
            '-' : b := b and (Tabla[j] <= Patron[j].Cantidad);
         end;
      if b then begin
         Inc(NumSol);
         if not Silencio then begin
            WriteLn(F, 'Solucion ', NumSol);
            for i := 1 to NPiezas do
               WriteLn(F, i:2, '-', Sol[i].Nombre, ': ', Sol[i].Posi, '. ', EscribeCoco(Sol[i].Desc[1]));
            WriteLn(F);
         end;
      end;
   until False;
   Close(F);
   Close(FichSol);
   ExtraeSol := NumSol;
end;

{
   Muestra un mensaje de error y aborta el programa.
}
procedure DiError(MsjErr : String);
begin
   WriteLn(MsjErr);
   Halt;
end;


{ -------------------------------------------------------------------------}


var
   P, L, Nombre : String;
   F : Text;
   i : Byte;
   t : Char;
   C : TStWord;
   Sol : TSolucion;
   NS, CodErr : Word;
   Patron : TPatron;


begin
   P := ParamStr(1);

   // Mostrar pieza que se pasa directamente en forma de cadena binaria
   if (P[1] = '0') or (P[1] = '1') then begin
      if Length(P) <> 16 then DiError('El argumento debe tener 16 dgitos de longitud');
      for i := 1 to 16 do
         if (P[i] <> '1') and (P[i] <> '0') then
            DiError('El argumento debe contener slo unos y ceros');
      MuestraCoco(P, '', 1, 0);
   end

   // Mostrar piezas de fichero de piezas
   else if Copy(P, 1, 2) = '-p' then begin
      P := ParamStr(2);
      Assign(F, P);
      {$i-}
      Reset(F);
      {$i+}
      if IOResult <> 0 then DiError('No encuentro el fichero de piezas.');
      t := 's';
      while not Eof(F) and (t <> 'q') and (t <> 'Q') do begin
         ReadLn(F, L);
         if (L = '') or (L[1] = ';') then Continue;
         i := Pos(' ', L);
         Nombre := Copy(L, 1, i-1);
         repeat
            Inc(i);
         until L[i] <> ' ';
         C := Copy(L, i, 16);
         t := MuestraCoco(C, Nombre, 1, 0);
      end;
      Close(F);
   end

   // Filtrar el fichero de soluciones eliminando soluciones equivalentes
   else if Copy(P, 1, 2) = '-f' then begin
      if ParamCount >= 3 then P := ParamStr(3) else P := 'figuras\cubo1x1x1.map';
      case LeeMapa(Mapa, P) of
         1 : DiError('No encuentro el fichero de mapa.');
         2 : DiError('Fichero de mapa incorrecto.');
      end;
      case FiltraSol(ParamStr(2)) of
         1 : DiError('No encuentro el fichero de soluciones.');
         2 : DiError('Rebasado el nmero mximo de soluciones.');
         4 : DiError('El fichero no tiene ninguna solucin.');
      end;
   end

   // Extraccin de soluciones segn tipos de piezas (colores o piezas concretas)
   else if Copy(P, 1, 2) = '-e' then begin
      FillByte(Patron, SizeOf(Patron), 0);
      for i := 3 to ParamCount do
         if (i mod 2) = 1 then Patron[(i-1) div 2].Color := ParamStr(i)
         else begin
            if (ParamStr(i)[1] >= '0') and (ParamStr(i)[1] <= '9') then begin
               Patron[(i-1) div 2].Op := '=';
               Val(ParamStr(i)[1], Patron[(i-1) div 2].Cantidad, CodErr);
            end else begin
               Patron[(i-1) div 2].Op := ParamStr(i)[1];
               Val(ParamStr(i)[2], Patron[(i-1) div 2].Cantidad, CodErr);
            end;
            if CodErr > 0 then DiError('Despus de cada tipo de pieza hay que escribir una cantidad');
         end;
      WriteLn(ExtraeSol(ParamStr(2), Patron, Copy(P, 1, 3) = '-es'), ' soluciones encontradas.');
   end

   // Mostrar solucin
   else if (Copy(P, 1, 2) = '-d') or (Copy(P, 1, 2) = '-s') then begin
      Val(ParamStr(3), NS, CodErr);
      if (CodErr > 0) or (NS = 0) then DiError('Hay que indicar un nmero de solucin');
      case LeeSol(Sol, ParamStr(2), NS) of
         1 : DiError('No encuentro el fichero de soluciones.');
         2 : DiError('No hay tantas soluciones en el fichero.');
      end;
      ExtiendeSol(Sol);
      if Copy(P, 1, 2) = '-d' then
         for i := 1 to NPiezas do begin
            with Sol[i] do t := MuestraCoco(EscribeCoco(Desc[1]), Nombre, Posi, i);
            if (t = 'q') or (t = 'Q') then Break;
         end
      else begin // -s
         if ParamCount >= 4 then P := ParamStr(4) else P := 'figuras\cubo1x1x1.map';
         case LeeMapa(Mapa, P) of
            1 : DiError('No encuentro el fichero de mapa.');
            2 : DiError('Fichero de mapa incorrecto.');
         end;
         VerSol(Sol);
      end;
   end

   // Informacin
   else begin
      WriteLn;
      WriteLn('Coco v. 1.0');
      WriteLn('Un programa complementario de CocoWirrel.');
      WriteLn;
      WriteLn('Modo de uso:');
      WriteLn;
      WriteLn('coco [cad_bin | -p f_pie | -d f_sol n | -s f_sol n [f_map] | -f f_sol [f_map] | -e[s] f_sol {t_pieza [[=]|+|-]n}]');
      WriteLn;
      WriteLn('cad_bin              Cadena binaria');
      WriteLn('f_pie                Fichero de piezas');
      WriteLn('f_sol                Fichero de soluciones');
      WriteLn('f_map                Fichero de mapa');
      WriteLn('n                    Nmero entero positivo (seguido de t_pieza puede ser 0)');
      WriteLn('t_pieza              Tipo de pieza');
      WriteLn;
      WriteLn('cad_bin              Muestra la pieza descrita por cad_bin');
      WriteLn('-p f_pie             Muestra las piezas de f_pie');
      WriteLn('-d f_sol n           Muestra las piezas de la solucin n de f_sol');
      WriteLn('-s f_sol n [f_map]   Muestra la solucin n de f_sol en modo figura');
      WriteLn('-f f_sol [f_map]     Filtra f_sol eliminando soluciones equivalentes');
      WriteLn('-e[s] f_sol {t_pieza [[=]|+|-]n} Extrae las soluciones de f_sol que tienen');
      WriteLn('                     ciertos tipos de piezas, por ejemplo azul 3 amarillo 3');
   end;
end.
